@@ -150,6 +150,9 @@
 // * Figure out best way to estimate completion time
 //   and display it in a ScanProgressMeter 
 
+#ifdef WIN32
+#include "nmap_winconfig.h"
+#endif
 
 #include <stdlib.h>
 #include <limits.h>
@@ -161,7 +164,7 @@
 #include "nmap_dns.h"
 #include "nsock.h"
 #include "utils.h"
-#include "tty.h"
+#include "nmap_tty.h"
 
 extern NmapOps o;
 
@@ -199,7 +202,7 @@ extern NmapOps o;
 
 // In milliseconds
 // Each row MUST be terminated with -1
-int read_timeouts[][4] = {
+static int read_timeouts[][4] = {
   { 4000, 4000, 5000, -1 }, // 1 server
   { 2500, 4000,   -1, -1 }, // 2 servers
   { 2500, 3000,   -1, -1 }, // 3+ servers
@@ -275,7 +278,6 @@ static std::list<request *> cname_reqs;
 static int total_reqs;
 static nsock_pool dnspool=NULL;
 
-static int etchosts_filled=0;
 static std::list<host_elem *> etchosts[HASH_TABLE_SIZE];
 
 static int stat_actual, stat_ok, stat_nx, stat_sf, stat_trans, stat_dropped, stat_cname;
@@ -288,10 +290,7 @@ static ScanProgressMeter *SPM;
 
 //------------------- Prototypes and macros ---------------------
 
-void close_dns_servers();
-void write_evt_handler(nsock_pool nsp, nsock_event evt, void *req_v);
-void do_possible_writes();
-int deal_with_timedout_reads();
+void put_dns_packet_on_wire(request *req);
 
 #define ACTION_FINISHED 0
 #define ACTION_CNAME_LIST 1
@@ -301,7 +300,7 @@ int deal_with_timedout_reads();
 
 //------------------- Misc code --------------------- 
 
-void output_summary() {
+static void output_summary() {
   int tp = stat_ok + stat_nx + stat_dropped;
   struct timeval now;
 
@@ -314,291 +313,30 @@ void output_summary() {
                     (unsigned long) servs.size(), stat_ok, stat_nx, stat_dropped, stat_sf, stat_trans);
 }
 
-
-void check_capacities(dns_server *tpserv) {
+static void check_capacities(dns_server *tpserv) {
   if (tpserv->capacity < CAPACITY_MIN) tpserv->capacity = CAPACITY_MIN;
   if (tpserv->capacity > CAPACITY_MAX) tpserv->capacity = CAPACITY_MAX;
   if (o.debugging >= TRACE_DEBUG_LEVEL) log_write(LOG_STDOUT, "CAPACITY <%s> = %d\n", tpserv->hostname, tpserv->capacity);
 }
 
+// Closes all nsis created in connect_dns_servers()
+static void close_dns_servers() {
+  std::list<dns_server *>::iterator serverI;
 
-
-
-//------------------- Read handling code ---------------------
-
-// After processing a DNS response, we search through the IPs we're
-// looking for and update their results as necessary.
-// Returns non-zero if this matches a query we're looking for
-int process_result(u32 ia, char *result, int action, u16 id) {
-  std::list<dns_server *>::iterator servI;
-  std::list<request *>::iterator reqI;
-  dns_server *tpserv;
-  request *tpreq;
-
-  for(servI = servs.begin(); servI != servs.end(); servI++) {
-    tpserv = *servI;
-
-    for(reqI = tpserv->in_process.begin(); reqI != tpserv->in_process.end(); reqI++) {
-      tpreq = *reqI;
-
-      if (id == tpreq->id) {
-
-        if (ia != 0 && (u32) (tpreq->targ->v4host().s_addr) != ia) continue;
-
-        if (action == ACTION_CNAME_LIST || action == ACTION_FINISHED) {
-        tpserv->capacity += CAPACITY_UP_STEP;
-        check_capacities(tpserv);
-
-        if (result) tpreq->targ->setHostName(result);
-        tpserv->in_process.remove(tpreq);
-        tpserv->reqs_on_wire--;
-
-        total_reqs--;
-
-          if (action == ACTION_CNAME_LIST) cname_reqs.push_back(tpreq);
-          if (action == ACTION_FINISHED) delete tpreq;
-        } else {
-          memcpy(&tpreq->timeout, nsock_gettimeofday(), sizeof(struct timeval));
-          deal_with_timedout_reads();
-        }
-
-        do_possible_writes();
-
-        // Close DNS servers if we're all done so that we kill
-        // all events and return from nsock_loop immediatley
-        if (total_reqs == 0)
-          close_dns_servers();
-        return 1;
-      }
+  for(serverI = servs.begin(); serverI != servs.end(); serverI++) {
+    if ((*serverI)->connected) {
+      nsi_delete((*serverI)->nsd, NSOCK_PENDING_SILENT);
+      (*serverI)->connected = 0;
+      (*serverI)->to_process.clear();
+      (*serverI)->in_process.clear();
     }
   }
-
-  return 0;
 }
 
 
-// Gets an IP address from a X.X.X.X.in-addr.arpa DNS
-// encoded string inside a packet.
-// maxlen is the very maximum length (in total bytes)
-// that should be processed
-u32 parse_inaddr_arpa(unsigned char *buf, int maxlen) {
-  u32 ip=0;
-  int i, j;
-
-  for (i=0; i<=3; i++) {
-    if (maxlen <= 0) return 0;
-
-    if (buf[0] < 1 || buf[0] > 3) return 0;
-
-    maxlen -= buf[0] + 1;
-    if (maxlen <= 0) return 0;
-
-    for (j=1; j<=buf[0]; j++) if (!isdigit(buf[j])) return 0;
-
-    ip |= atoi((char *) buf+1) << (8*i);
-    buf += buf[0] + 1;
-  }
-
-  if (maxlen < 14) return 0; // length of the following string
-  if (strcasecmp((char *) buf, "\x07in-addr\004arpa\0")) return 0;
-
-  return ntohl(ip);
-}
-
-
-// Turns a DNS packet encoded name (see the RFC) and turns it into
-// a normal decimal separated hostname.
-// ASSUMES NAME LENGTH/VALIDITY HAS ALREADY BEEN VERIFIED
-int encoded_name_to_normal(unsigned char *buf, char *output, int outputsize){
-  while (buf[0]) {
-    if (buf[0] >= outputsize-1) return -1;
-    memcpy(output, buf+1, buf[0]);
-    outputsize -= buf[0];
-    output += buf[0];
-    buf += buf[0]+1;
-
-    if (buf[0]) {
-      *output++ = '.';
-      outputsize--;
-    } else {
-      *output = '\0';
-    }
-  }
-
-  return 0;
-}
-
-
-// Takes a pointer to the start of a DNS name inside a packet. It makes
-// sure that there is enough space in the name, deals with compression, etc.
-int advance_past_dns_name(unsigned char *buf, int buflen, int curbuf, 
-			  int *nameloc) {
-  int compression=0;
-
-  if (curbuf <= 0 || curbuf >= buflen) return -1;
-
-  if ((buf[curbuf] & 0xc0)) {
-    // Need 2 bytes for compression info
-    if (curbuf + 1 >= buflen) return -1;
-
-    // Compression is OK
-    compression = curbuf+2;
-    curbuf = ((buf[curbuf+1] & 0xFF) + ((buf[curbuf] & 0xFF) << 8)) & 0x3FFF;
-    if (curbuf < 0 || curbuf >= buflen) return -1;
-  }
-
-  if (nameloc != NULL) *nameloc = curbuf;
-
-  while(buf[curbuf]) {
-    if (curbuf + buf[curbuf] >= buflen || buf[curbuf] <= 0) return -1;
-    curbuf += buf[curbuf] + 1;
-  }
-
-  if (compression) return compression;
-  else return curbuf+1;
-}
-
-
-// Nsock read handler. One nsock read for each DNS server exists at each
-// time. This function uses various helper functions as defined above.
-void read_evt_handler(nsock_pool nsp, nsock_event evt, void *nothing) {
-  unsigned char *buf;
-  int buflen, curbuf=0;
-  int i, nameloc, rdlen, atype, aclass;
-  int errcode=0;
-  int queries, answers;
-  u16 packet_id;
-
-  if (total_reqs >= 1)
-    nsock_read(nsp, nse_iod(evt), read_evt_handler, -1, NULL);
-
-  if (nse_type(evt) != NSE_TYPE_READ || nse_status(evt) != NSE_STATUS_SUCCESS) {
-    if (o.debugging)
-      log_write(LOG_STDOUT, "mass_dns: warning: got a %s:%s in read_evt_handler()\n",
-        nse_type2str(nse_type(evt)),
-        nse_status2str(nse_status(evt)));
-    return;
-  }
-
-  buf = (unsigned char *) nse_readbuf(evt, &buflen);
-
-  // Size of header is 12, and we must have additional data as well
-  if (buflen <= 12) return;
-
-  packet_id = (buf[1] & 0xFF) + ((buf[0] & 0xFF) << 8);
-
-  // Check that this is a response, standard query, and that no truncation was performed
-  // 0xFA == 11111010 (we're not concerned with AA or RD bits)
-  if ((buf[2] & 0xFA) != 0x80) return;
-
-  // Check that Recursion is available, the zero field is all zeros
-  // and there is no error condition:
-  if ((buf[3] & 0xFF) != 0x80) {
-    if ((buf[3] & 0xF) == 2) errcode = 2;
-    else if ((buf[3] & 0xF) == 3) errcode = 3;
-    else return;
-  }
-
-  queries = (buf[5] & 0xFF) + ((buf[4] & 0xFF) << 8);
-  answers = (buf[7] & 0xFF) + ((buf[6] & 0xFF) << 8);
-
-  // With a normal resolution, we should have 1+ queries and 1+ answers.
-  // If the domain doesn't resolve (NXDOMAIN or SERVFAIL) we should have
-  // 1+ queries and 0 answers:
-  if (errcode) {
-    int found;
-
-    // NXDomain means we're finished (doesn't exist for sure)
-    // but SERVFAIL might just mean a server timeout
-    found = process_result(0, NULL, errcode == 3 ? ACTION_FINISHED : ACTION_TIMEOUT, packet_id);
-
-    if (errcode == 2 && found) {
-      if (o.debugging >= TRACE_DEBUG_LEVEL) log_write(LOG_STDOUT, "mass_rdns: SERVFAIL <id = %d>\n", packet_id);
-      stat_sf++;
-    } else if (errcode == 3 && found) {
-      if (o.debugging >= TRACE_DEBUG_LEVEL) log_write(LOG_STDOUT, "mass_rdns: NXDOMAIN <id = %d>\n", packet_id);
-      output_summary();
-      stat_nx++;
-  }
-
-    return;
-  }
-
-  if (queries <= 0 || answers <= 0) return;
-
-  curbuf = 12;
-
-  // Need to safely skip past QUERY section
-
-  for (i=0; i<queries; i++) {
-    curbuf = advance_past_dns_name(buf, buflen, curbuf, &nameloc);
-    if (curbuf == -1) return;
-
-    // Make sure we have the QTYPE and QCLASS fields
-    if (curbuf + 4 >= buflen) return;
-    curbuf += 4;
-      }
-
-  // We're now at the ANSWER section
-
-  for (i=0; i<answers; i++) {
-    curbuf = advance_past_dns_name(buf, buflen, curbuf, &nameloc);
-    if (curbuf == -1) return;
-
-    // Make sure we have the TYPE (2), CLASS (2), TTL (4), and
-    // RDLENGTH (2) fields
-    if (curbuf + 10 >= buflen) return;
-
-    atype = (buf[curbuf+1] & 0xFF) + ((buf[curbuf+0] & 0xFF) << 8);
-    aclass = (buf[curbuf+3] & 0xFF) + ((buf[curbuf+2] & 0xFF) << 8);
-    rdlen = (buf[curbuf+9] & 0xFF) + ((buf[curbuf+8] & 0xFF) << 8);
-    curbuf += 10;
-
-    if (atype == 12 && aclass == 1) {
-      // TYPE 12 is PTR
-      struct in_addr ia;
-      char outbuf[512];
-
-      ia.s_addr = parse_inaddr_arpa(buf+nameloc, buflen-nameloc);
-      if (ia.s_addr == 0) return;
-
-      curbuf = advance_past_dns_name(buf, buflen, curbuf, &nameloc);
-      if (curbuf == -1 || curbuf > buflen) return;
-
-      if (encoded_name_to_normal(buf+nameloc, outbuf, sizeof(outbuf)) == -1) return;
-
-      if (process_result(ia.s_addr, outbuf, ACTION_FINISHED, packet_id)) {
-        if (o.debugging >= TRACE_DEBUG_LEVEL) log_write(LOG_STDOUT, "mass_rdns: OK MATCHED <%s> to <%s>\n", inet_ntoa(ia), outbuf);
-        output_summary();
-        stat_ok++;
-      }
-    } else if (atype == 5 && aclass == 1) {
-      // TYPE 5 is CNAME
-      struct in_addr ia;
-
-      ia.s_addr = parse_inaddr_arpa(buf+nameloc, buflen-nameloc);
-      if (ia.s_addr == 0) return;
-
-      if (o.debugging >= TRACE_DEBUG_LEVEL) log_write(LOG_STDOUT, "mass_rdns: CNAME found for <%s>\n", inet_ntoa(ia));
-      process_result(ia.s_addr, NULL, ACTION_CNAME_LIST, packet_id);
-    } else {
-      if (rdlen < 0 || rdlen + curbuf >= buflen) return;
-      curbuf += rdlen;
-    }
-
-    if (curbuf >= buflen) return;
-  }
-
-}
-
-
-
-
-//------------------- Write handling code ---------------------
-
 // Inserts an integer (endian non-specifically) into a DNS packet.
 // Returns number of bytes written
-int add_integer_to_dns_packet(char *packet, int c) {
+static int add_integer_to_dns_packet(char *packet, int c) {
   char tpnum[4];
   int tplen;
 
@@ -610,6 +348,44 @@ int add_integer_to_dns_packet(char *packet, int c) {
   return tplen+1;
 }
 
+// Puts as many packets on the line as capacity will allow
+static void do_possible_writes() {
+  std::list<dns_server *>::iterator servI;
+  dns_server *tpserv;
+  request *tpreq;
+
+  for(servI = servs.begin(); servI != servs.end(); servI++) {
+    tpserv = *servI;
+
+    if (tpserv->write_busy == 0 && tpserv->reqs_on_wire < tpserv->capacity) {
+      tpreq = NULL;
+      if (!tpserv->to_process.empty()) {
+        tpreq = tpserv->to_process.front();
+        tpserv->to_process.pop_front();
+      } else if (!new_reqs.empty()) {
+        tpreq = new_reqs.front();
+        tpreq->first_server = tpreq->curr_server = tpserv;
+        new_reqs.pop_front();
+      }
+
+      if (tpreq) {
+        if (o.debugging >= TRACE_DEBUG_LEVEL) log_write(LOG_STDOUT, "mass_rdns: TRANSMITTING for <%s> (server <%s>)\n", tpreq->targ->targetipstr() , tpserv->hostname);
+        stat_trans++;
+        put_dns_packet_on_wire(tpreq);
+      }
+    }
+  }
+}
+
+// nsock write handler
+static void write_evt_handler(nsock_pool nsp, nsock_event evt, void *req_v) {
+  request *req = (request *) req_v;
+
+  req->curr_server->write_busy = 0;
+  req->curr_server->in_process.push_front(req);
+
+  do_possible_writes();
+}
 
 // Takes a DNS request structure and actually puts it on the wire
 // (calls nsock_write()). Does various other tasks like recording
@@ -649,12 +425,9 @@ void put_dns_packet_on_wire(request *req) {
   nsock_write(dnspool, req->curr_server->nsd, write_evt_handler, WRITE_TIMEOUT, req, packet, plen);
 }
 
-
-
-
 // Processes DNS packets that have timed out
 // Returns time until next read timeout
-int deal_with_timedout_reads() {
+static int deal_with_timedout_reads() {
   std::list<dns_server *>::iterator servI;
   std::list<dns_server *>::iterator servItemp;
   std::list<request *>::iterator reqI;
@@ -736,61 +509,279 @@ int deal_with_timedout_reads() {
 
 }
 
-
-// Puts as many packets on the line as capacity will allow
-void do_possible_writes() {
+// After processing a DNS response, we search through the IPs we're
+// looking for and update their results as necessary.
+// Returns non-zero if this matches a query we're looking for
+static int process_result(u32 ia, char *result, int action, u16 id) {
   std::list<dns_server *>::iterator servI;
+  std::list<request *>::iterator reqI;
   dns_server *tpserv;
   request *tpreq;
 
   for(servI = servs.begin(); servI != servs.end(); servI++) {
     tpserv = *servI;
 
-    if (tpserv->write_busy == 0 && tpserv->reqs_on_wire < tpserv->capacity) {
-      tpreq = NULL;
-      if (!tpserv->to_process.empty()) {
-        tpreq = tpserv->to_process.front();
-        tpserv->to_process.pop_front();
-      } else if (!new_reqs.empty()) {
-        tpreq = new_reqs.front();
-        tpreq->first_server = tpreq->curr_server = tpserv;
-        new_reqs.pop_front();
-      }
+    for(reqI = tpserv->in_process.begin(); reqI != tpserv->in_process.end(); reqI++) {
+      tpreq = *reqI;
 
-      if (tpreq) {
-        if (o.debugging >= TRACE_DEBUG_LEVEL) log_write(LOG_STDOUT, "mass_rdns: TRANSMITTING for <%s> (server <%s>)\n", tpreq->targ->targetipstr() , tpserv->hostname);
-        stat_trans++;
-        put_dns_packet_on_wire(tpreq);
+      if (id == tpreq->id) {
+
+        if (ia != 0 && (u32) (tpreq->targ->v4host().s_addr) != ia) continue;
+
+        if (action == ACTION_CNAME_LIST || action == ACTION_FINISHED) {
+        tpserv->capacity += CAPACITY_UP_STEP;
+        check_capacities(tpserv);
+
+        if (result) tpreq->targ->setHostName(result);
+        tpserv->in_process.remove(tpreq);
+        tpserv->reqs_on_wire--;
+
+        total_reqs--;
+
+          if (action == ACTION_CNAME_LIST) cname_reqs.push_back(tpreq);
+          if (action == ACTION_FINISHED) delete tpreq;
+        } else {
+          memcpy(&tpreq->timeout, nsock_gettimeofday(), sizeof(struct timeval));
+          deal_with_timedout_reads();
+        }
+
+        do_possible_writes();
+
+        // Close DNS servers if we're all done so that we kill
+        // all events and return from nsock_loop immediatley
+        if (total_reqs == 0)
+          close_dns_servers();
+        return 1;
       }
     }
+  }
 
+  return 0;
+}
+
+
+// Gets an IP address from a X.X.X.X.in-addr.arpa DNS
+// encoded string inside a packet.
+// maxlen is the very maximum length (in total bytes)
+// that should be processed
+static u32 parse_inaddr_arpa(unsigned char *buf, int maxlen) {
+  u32 ip=0;
+  int i, j;
+
+  for (i=0; i<=3; i++) {
+    if (maxlen <= 0) return 0;
+
+    if (buf[0] < 1 || buf[0] > 3) return 0;
+
+    maxlen -= buf[0] + 1;
+    if (maxlen <= 0) return 0;
+
+    for (j=1; j<=buf[0]; j++) if (!isdigit(buf[j])) return 0;
+
+    ip |= atoi((char *) buf+1) << (8*i);
+    buf += buf[0] + 1;
+  }
+
+  if (maxlen < 14) return 0; // length of the following string
+  if (strcasecmp((char *) buf, "\x07in-addr\004arpa\0")) return 0;
+
+  return ntohl(ip);
+}
+
+
+// Turns a DNS packet encoded name (see the RFC) and turns it into
+// a normal decimal separated hostname.
+// ASSUMES NAME LENGTH/VALIDITY HAS ALREADY BEEN VERIFIED
+int encoded_name_to_normal(unsigned char *buf, char *output, int outputsize){
+  while (buf[0]) {
+    if (buf[0] >= outputsize-1) return -1;
+    memcpy(output, buf+1, buf[0]);
+    outputsize -= buf[0];
+    output += buf[0];
+    buf += buf[0]+1;
+
+    if (buf[0]) {
+      *output++ = '.';
+      outputsize--;
+    } else {
+      *output = '\0';
+    }
+  }
+
+  return 0;
+}
+
+
+// Takes a pointer to the start of a DNS name inside a packet. It makes
+// sure that there is enough space in the name, deals with compression, etc.
+static int advance_past_dns_name(u8 *buf, int buflen, int curbuf, 
+			  int *nameloc) {
+  int compression=0;
+
+  if (curbuf <= 0 || curbuf >= buflen) return -1;
+
+  if ((buf[curbuf] & 0xc0)) {
+    // Need 2 bytes for compression info
+    if (curbuf + 1 >= buflen) return -1;
+
+    // Compression is OK
+    compression = curbuf+2;
+    curbuf = (buf[curbuf+1] + (buf[curbuf] << 8)) & 0x3FFF;
+    if (curbuf < 0 || curbuf >= buflen) return -1;
+  }
+
+  if (nameloc != NULL) *nameloc = curbuf;
+
+  while(buf[curbuf]) {
+    if (curbuf + buf[curbuf] >= buflen || buf[curbuf] <= 0) return -1;
+    curbuf += buf[curbuf] + 1;
+  }
+
+  if (compression) return compression;
+  else return curbuf+1;
+}
+
+// Nsock read handler. One nsock read for each DNS server exists at each
+// time. This function uses various helper functions as defined above.
+static void read_evt_handler(nsock_pool nsp, nsock_event evt, void *nothing) {
+  u8 *buf;
+  int buflen, curbuf=0;
+  int i, nameloc, rdlen, atype, aclass;
+  int errcode=0;
+  int queries, answers;
+  u16 packet_id;
+
+  if (total_reqs >= 1)
+    nsock_read(nsp, nse_iod(evt), read_evt_handler, -1, NULL);
+
+  if (nse_type(evt) != NSE_TYPE_READ || nse_status(evt) != NSE_STATUS_SUCCESS) {
+    if (o.debugging)
+      log_write(LOG_STDOUT, "mass_dns: warning: got a %s:%s in read_evt_handler()\n",
+        nse_type2str(nse_type(evt)),
+        nse_status2str(nse_status(evt)));
+    return;
+  }
+
+  buf = (unsigned char *) nse_readbuf(evt, &buflen);
+
+  // Size of header is 12, and we must have additional data as well
+  if (buflen <= 12) return;
+
+  packet_id = buf[1] + (buf[0] << 8);
+
+  // Check that this is a response, standard query, and that no truncation was performed
+  // 0xFA == 11111010 (we're not concerned with AA or RD bits)
+  if ((buf[2] & 0xFA) != 0x80) return;
+
+  // Check that Recursion is available, the zero field is all zeros
+  // and there is no error condition:
+  if (buf[3] != 0x80) {
+    if ((buf[3] & 0xF) == 2) errcode = 2;
+    else if ((buf[3] & 0xF) == 3) errcode = 3;
+    else return;
+  }
+
+  queries = buf[5] + (buf[4] << 8);
+  answers = buf[7] + (buf[6] << 8);
+
+  // With a normal resolution, we should have 1+ queries and 1+ answers.
+  // If the domain doesn't resolve (NXDOMAIN or SERVFAIL) we should have
+  // 1+ queries and 0 answers:
+  if (errcode) {
+    int found;
+
+    // NXDomain means we're finished (doesn't exist for sure)
+    // but SERVFAIL might just mean a server timeout
+    found = process_result(0, NULL, errcode == 3 ? ACTION_FINISHED : ACTION_TIMEOUT, packet_id);
+
+    if (errcode == 2 && found) {
+      if (o.debugging >= TRACE_DEBUG_LEVEL) log_write(LOG_STDOUT, "mass_rdns: SERVFAIL <id = %d>\n", packet_id);
+      stat_sf++;
+    } else if (errcode == 3 && found) {
+      if (o.debugging >= TRACE_DEBUG_LEVEL) log_write(LOG_STDOUT, "mass_rdns: NXDOMAIN <id = %d>\n", packet_id);
+      output_summary();
+      stat_nx++;
+  }
+
+    return;
+  }
+
+  if (queries <= 0 || answers <= 0) return;
+
+  curbuf = 12;
+
+  // Need to safely skip past QUERY section
+
+  for (i=0; i<queries; i++) {
+    curbuf = advance_past_dns_name(buf, buflen, curbuf, &nameloc);
+    if (curbuf == -1) return;
+
+    // Make sure we have the QTYPE and QCLASS fields
+    if (curbuf + 4 >= buflen) return;
+    curbuf += 4;
+      }
+
+  // We're now at the ANSWER section
+
+  for (i=0; i<answers; i++) {
+    curbuf = advance_past_dns_name(buf, buflen, curbuf, &nameloc);
+    if (curbuf == -1) return;
+
+    // Make sure we have the TYPE (2), CLASS (2), TTL (4), and
+    // RDLENGTH (2) fields
+    if (curbuf + 10 >= buflen) return;
+
+    atype = buf[curbuf+1] + (buf[curbuf+0] << 8);
+    aclass = buf[curbuf+3] + (buf[curbuf+2] << 8);
+    rdlen = buf[curbuf+9] + (buf[curbuf+8] << 8);
+    curbuf += 10;
+
+    if (atype == 12 && aclass == 1) {
+      // TYPE 12 is PTR
+      struct in_addr ia;
+      char outbuf[512];
+
+      ia.s_addr = parse_inaddr_arpa(buf+nameloc, buflen-nameloc);
+      if (ia.s_addr == 0) return;
+
+      curbuf = advance_past_dns_name(buf, buflen, curbuf, &nameloc);
+      if (curbuf == -1 || curbuf > buflen) return;
+
+      if (encoded_name_to_normal(buf+nameloc, outbuf, sizeof(outbuf)) == -1) return;
+
+      if (process_result(ia.s_addr, outbuf, ACTION_FINISHED, packet_id)) {
+        if (o.debugging >= TRACE_DEBUG_LEVEL) log_write(LOG_STDOUT, "mass_rdns: OK MATCHED <%s> to <%s>\n", inet_ntoa(ia), outbuf);
+        output_summary();
+        stat_ok++;
+      }
+    } else if (atype == 5 && aclass == 1) {
+      // TYPE 5 is CNAME
+      struct in_addr ia;
+
+      ia.s_addr = parse_inaddr_arpa(buf+nameloc, buflen-nameloc);
+      if (ia.s_addr == 0) return;
+
+      if (o.debugging >= TRACE_DEBUG_LEVEL) log_write(LOG_STDOUT, "mass_rdns: CNAME found for <%s>\n", inet_ntoa(ia));
+      process_result(ia.s_addr, NULL, ACTION_CNAME_LIST, packet_id);
+    } else {
+      if (rdlen < 0 || rdlen + curbuf >= buflen) return;
+      curbuf += rdlen;
+    }
+
+    if (curbuf >= buflen) return;
   }
 
 }
 
 
-// nsock write handler
-void write_evt_handler(nsock_pool nsp, nsock_event evt, void *req_v) {
-  request *req = (request *) req_v;
-
-  req->curr_server->write_busy = 0;
-  req->curr_server->in_process.push_front(req);
-
-  do_possible_writes();
-}
-  
-
-
-//------------------- DNS Server handling code ---------------------
-
 // nsock connect handler - Empty because it doesn't really need to do anything...
-void connect_evt_handler(nsock_pool nsp, nsock_event evt, void *servers) {
+static void connect_evt_handler(nsock_pool nsp, nsock_event evt, void *servers) {
 }
 
 
 // Adds DNS servers to the dns_server list. They can be separated by
 // commas or spaces - NOTE this doesn't actually do any connecting!
-void add_dns_server(char *ipaddrs) {
+static void add_dns_server(char *ipaddrs) {
   std::list<dns_server *>::iterator servI;
   dns_server *tpserv;
   char *hostname;
@@ -846,21 +837,6 @@ void connect_dns_servers() {
 }
 
 
-// Closes all nsis created in connect_dns_servers()
-void close_dns_servers() {
-  std::list<dns_server *>::iterator serverI;
-
-  for(serverI = servs.begin(); serverI != servs.end(); serverI++) {
-    if ((*serverI)->connected) {
-      nsi_delete((*serverI)->nsd, NSOCK_PENDING_SILENT);
-      (*serverI)->connected = 0;
-      (*serverI)->to_process.clear();
-      (*serverI)->in_process.clear();
-    }
-  }
-
-}
-
 #ifdef WIN32
 void win32_read_registry(char *controlset) {
   HKEY hKey;
@@ -921,7 +897,7 @@ void win32_read_registry(char *controlset) {
 
 // Parses /etc/resolv.conf (unix) or the registry (win32) and adds
 // all the nameservers found via the add_dns_server() function.
-void parse_resolvdotconf() {
+static void parse_resolvdotconf() {
 
 #ifndef WIN32
 
@@ -956,7 +932,7 @@ void parse_resolvdotconf() {
 }
 
 
-void parse_etchosts(char *fname) {
+static void parse_etchosts(char *fname) {
   FILE *fp;
   char buf[2048], hname[256], ipaddrstr[16], *tp;
   struct in_addr ia;
@@ -990,7 +966,7 @@ void parse_etchosts(char *fname) {
 }
 
 
-char *lookup_etchosts(u32 ip) {
+static char *lookup_etchosts(u32 ip) {
   std::list<host_elem *>::iterator hostI;
   host_elem *tpelem;
 
@@ -1003,13 +979,41 @@ char *lookup_etchosts(u32 ip) {
 }
 
 
+static void etchosts_init(void) {
+  static int initialized = 0;
+  if (initialized) return;
+  initialized = 1;
+
+#ifdef WIN32
+  char windows_dir[1024];
+  char tpbuf[2048];
+  int has_backslash;
+
+  if (!GetWindowsDirectory(windows_dir, sizeof(windows_dir)))
+    fatal("Failed to determine your windows directory");
+
+  // If it has a backslash it's C:\, otherwise something like C:\WINNT
+  has_backslash = (windows_dir[strlen(windows_dir)-1] == '\\');
+
+  // Windows 95/98/Me:
+  snprintf(tpbuf, sizeof(tpbuf), "%s%shosts", windows_dir, has_backslash ? "" : "\\");
+  parse_etchosts(tpbuf);
+
+  // Windows NT/2000/XP/2K3:
+  snprintf(tpbuf, sizeof(tpbuf), "%s%ssystem32\\drivers\\etc\\hosts", windows_dir, has_backslash ? "" : "\\");
+  parse_etchosts(tpbuf);
+
+#else
+  parse_etchosts("/etc/hosts");
+#endif
+}
 
 
 //------------------- Main loops ---------------------
 
 
 // Actual main loop
-void nmap_mass_rdns_core(Target **targets, int num_targets) {
+static void nmap_mass_rdns_core(Target **targets, int num_targets) {
 
   Target **hostI;
   std::list<request *>::iterator reqI;
@@ -1066,32 +1070,7 @@ void nmap_mass_rdns_core(Target **targets, int num_targets) {
 
 
   // If necessary, set up the /etc/hosts hashtable
-  if (etchosts_filled == 0) {
-    #ifdef WIN32
-    char windows_dir[1024];
-    char tpbuf[2048];
-    int has_backslash;
-
-    if (!GetWindowsDirectory(windows_dir, sizeof(windows_dir)))
-      fatal("Failed to determine your windows directory");
-
-    // If it has a backslash it's C:\, otherwise something like C:\WINNT
-    has_backslash = (windows_dir[strlen(windows_dir)-1] == '\\');
-
-    // Windows 95/98/Me:
-    snprintf(tpbuf, sizeof(tpbuf), "%s%shosts", windows_dir, has_backslash ? "" : "\\");
-    parse_etchosts(tpbuf);
-
-    // Windows NT/2000/XP/2K3:
-    snprintf(tpbuf, sizeof(tpbuf), "%s%ssystem32\\drivers\\etc\\hosts", windows_dir, has_backslash ? "" : "\\");
-    parse_etchosts(tpbuf);
-
-    #else
-    parse_etchosts("/etc/hosts");
-    #endif
-
-    etchosts_filled = 1;
-  }
+  etchosts_init();
 
 
   total_reqs = 0;
